@@ -387,6 +387,7 @@ gail_finish_select (GtkWidget *widget)
           void *vp_next_focus_widget = &next_focus_widget;
           g_source_remove (focus_notify_handler);
           g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), vp_next_focus_widget);
+	  next_focus_widget = NULL;
           focus_notify_handler = 0;
           was_deselect = FALSE;
         }
@@ -569,7 +570,12 @@ gail_focus_notify (GtkWidget *widget)
         atk_obj  = gail_get_accessible_for_widget (focus_widget, &transient);
       else
         atk_obj = NULL;
-      atk_focus_tracker_notify (atk_obj);
+      /*
+       * Do not report focus on redundant object
+       */
+      if (atk_obj && 
+	  (atk_object_get_role(atk_obj) != ATK_ROLE_REDUNDANT_OBJECT))
+	  atk_focus_tracker_notify (atk_obj);
       if (atk_obj && transient)
         g_object_unref (atk_obj);
       if (subsequent_focus_widget)
@@ -613,6 +619,7 @@ gail_focus_notify_when_idle (GtkWidget *widget)
 	    {
 	      void *vp_next_focus_widget = &next_focus_widget;
 	      g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), vp_next_focus_widget);
+	      next_focus_widget = NULL;
 	    }
         }
       else
@@ -678,6 +685,7 @@ gail_deactivate_watcher (GSignalInvocationHint *ihint,
       void *vp_next_focus_widget = &next_focus_widget;
       g_source_remove (focus_notify_handler);
       g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), vp_next_focus_widget);
+      next_focus_widget = NULL;
       focus_notify_handler = 0;
       was_deselect = FALSE;
     }
@@ -757,7 +765,11 @@ gail_focus_object_destroyed (gpointer data)
 static void
 gail_focus_tracker (AtkObject *focus_object)
 {
-  if (focus_object)
+  /*
+   * Do not report focus on redundant object
+   */
+  if (focus_object && 
+      (atk_object_get_role(focus_object) != ATK_ROLE_REDUNDANT_OBJECT))
     {
       AtkObject *old_focus_object;
 
