@@ -64,6 +64,10 @@ static G_CONST_RETURN gchar* gail_button_action_get_name(AtkAction      *action,
 static gboolean              gail_button_set_description(AtkAction      *action,
                                                          gint           i,
                                                          const gchar    *desc);
+static void                  gail_button_notify_label_weak_ref (gpointer data,
+                                                                GObject  *obj);
+static void                  gail_button_notify_weak_ref       (gpointer data,
+                                                                GObject  *obj);
 
 
 /* AtkImage.h */
@@ -289,6 +293,34 @@ gail_button_get_name (AtkObject *obj)
   return name;
 }
 
+/*
+ * A DownArrow in a GtkToggltButton whose parent is not a ColorCombo
+ * has press as default action.
+ */
+static gboolean
+gail_button_is_default_press (GtkWidget *widget)
+{
+  GtkWidget  *child;
+  GtkWidget  *parent;
+  gboolean ret = FALSE;
+  const gchar *parent_type_name;
+
+  child = GTK_BIN (widget)->child;
+  if (GTK_IS_ARROW (child) &&
+      GTK_ARROW (child)->arrow_type == GTK_ARROW_DOWN)
+    {
+      parent = gtk_widget_get_parent (widget);
+      if (parent)
+        {
+          parent_type_name = g_type_name (G_OBJECT_TYPE (parent));
+          if (strcmp (parent_type_name, "ColorCombo"))
+            return TRUE;
+        }
+    }
+
+  return ret;
+}
+
 static void
 gail_button_real_initialize (AtkObject *obj,
                              gpointer   data)
@@ -296,7 +328,6 @@ gail_button_real_initialize (AtkObject *obj,
   GailButton *button = GAIL_BUTTON (obj);
   GtkWidget  *label;
   GtkWidget  *widget;
-  GtkWidget  *child;
 
   ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
 
@@ -332,15 +363,9 @@ gail_button_real_initialize (AtkObject *obj,
                           G_CALLBACK (gail_button_label_map_gtk),
                           button);
     }
-  child = GTK_BIN (widget)->child;
-  if (GTK_IS_ARROW (child) &&
-      GTK_ARROW (child)->arrow_type == GTK_ARROW_DOWN)
-    button->is_down_arrow = TRUE;
-  else
-    button->is_down_arrow = FALSE;
+  button->default_is_press = gail_button_is_default_press (widget);
     
   set_role_for_button (obj, data);
-
 }
 
 static void
@@ -387,6 +412,44 @@ gail_button_notify_label_gtk (GObject           *obj,
     }
 }
 
+static void
+gail_button_notify_weak_ref (gpointer data, GObject* obj)
+{
+  GtkLabel *label = NULL;
+
+  AtkObject* atk_obj = ATK_OBJECT (obj);
+  if (data && GTK_IS_WIDGET (data))
+    {
+      label = GTK_LABEL (data);
+      if (label)
+        {
+          g_signal_handlers_disconnect_by_func (label,
+                                                (GCallback) gail_button_notify_label_gtk,
+                                                GAIL_BUTTON (atk_obj));
+          g_object_weak_unref (G_OBJECT (label),
+                               gail_button_notify_label_weak_ref,
+                               GAIL_BUTTON (atk_obj));
+        }
+    }
+}
+
+static void
+gail_button_notify_label_weak_ref (gpointer data, GObject* obj)
+{
+  GtkLabel *label = NULL;
+  GailButton *button = NULL;
+
+  label = GTK_LABEL (obj);
+  if (data && GAIL_IS_BUTTON (data))
+    {
+      button = GAIL_BUTTON (ATK_OBJECT (data));
+      if (button)
+        g_object_weak_unref (G_OBJECT (button), gail_button_notify_weak_ref,
+                             label);
+    }
+}
+
+
 static void
 gail_button_init_textutil (GailButton  *button,
                            GtkWidget   *label)
@@ -396,6 +459,10 @@ gail_button_init_textutil (GailButton  *button,
   button->textutil = gail_text_util_new ();
   label_text = gtk_label_get_text (GTK_LABEL (label));
   gail_text_util_text_setup (button->textutil, label_text);
+  g_object_weak_ref (G_OBJECT (button),
+                     gail_button_notify_weak_ref, label);
+  g_object_weak_ref (G_OBJECT (label),
+                     gail_button_notify_label_weak_ref, button);
   g_signal_connect (label,
                     "notify",
                     (GCallback) gail_button_notify_label_gtk,
@@ -489,25 +556,35 @@ idle_do_action (gpointer data)
   GtkButton *button; 
   GtkWidget *widget;
   GailButton *gail_button;
+  GdkEvent tmp_event;
+
+  GDK_THREADS_ENTER ();
 
   gail_button = GAIL_BUTTON (data);
   gail_button->action_idle_handler = 0;
-
+  tmp_event.button.type = GDK_BUTTON_RELEASE;
+  tmp_event.button.window = widget->window;
+  tmp_event.button.button = 1;
+  tmp_event.button.send_event = TRUE;
+  tmp_event.button.time = GDK_CURRENT_TIME;
+  tmp_event.button.axes = NULL;
+  
   widget = GTK_ACCESSIBLE (gail_button)->widget;
-  if (widget == NULL)
-    /*
-     * State is defunct
-     */
-    return FALSE;
 
-  if (!GTK_WIDGET_IS_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
-    return FALSE;
+  if (widget == NULL /* State is defunct */ ||
+      !GTK_WIDGET_IS_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
+    {
+      GDK_THREADS_LEAVE ();
+      return FALSE;
+    }
+  else
+    gtk_widget_event (widget, &tmp_event);
 
   button = GTK_BUTTON (widget); 
   while (!g_queue_is_empty (gail_button->action_queue)) 
     {
       gint action_number = (gint) g_queue_pop_head (gail_button->action_queue);
-      if (gail_button->is_down_arrow)
+      if (gail_button->default_is_press)
         {
           if (action_number == 0)
             action_number = 1;
@@ -520,35 +597,33 @@ idle_do_action (gpointer data)
 	  gtk_widget_activate (widget);
 	  break;
 	case 1:
-	  {
-	    GdkEvent tmp_event;
-	    
-	    button->in_button = TRUE;
-	    gtk_button_enter (button);
-	    /*
-	     * Simulate a button press event. calling gtk_button_pressed() does
-	     * not get the job done for a GtkOptionMenu.  
-	     */
-	    tmp_event.button.type = GDK_BUTTON_PRESS;
-	    tmp_event.button.window = widget->window;
-	    tmp_event.button.button = 1;
-	    tmp_event.button.send_event = TRUE;
-	    tmp_event.button.time = GDK_CURRENT_TIME;
-	    tmp_event.button.axes = NULL;
-	    
-	    gtk_widget_event (widget, &tmp_event);
-	    break;
-	  }
+	  button->in_button = TRUE;
+	  gtk_button_enter (button);
+	  /*
+	   * Simulate a button press event. calling gtk_button_pressed() does
+	   * not get the job done for a GtkOptionMenu.  
+	   */
+	  tmp_event.button.type = GDK_BUTTON_PRESS;
+	  tmp_event.button.window = widget->window;
+	  tmp_event.button.button = 1;
+	  tmp_event.button.send_event = TRUE;
+	  tmp_event.button.time = GDK_CURRENT_TIME;
+	  tmp_event.button.axes = NULL;
+	  
+	  gtk_widget_event (widget, &tmp_event);
+	  break;
 	case 2:
 	  button->in_button = FALSE;
 	  gtk_button_leave (button);
-	  gtk_button_released (button);
 	  break;
 	default:
 	  g_assert_not_reached ();
 	  break;
 	}
     }
+
+  GDK_THREADS_LEAVE ();
+
   return FALSE; 
 }
 
@@ -567,7 +642,7 @@ gail_button_get_description (AtkAction *action,
 
   button = GAIL_BUTTON (action);
 
-  if (button->is_down_arrow)
+  if (button->default_is_press)
     {
       if (i == 0)
         i = 1;
@@ -600,7 +675,7 @@ gail_button_get_keybinding (AtkAction *action,
   gchar *return_value = NULL;
 
   button = GAIL_BUTTON (action);
-  if (button->is_down_arrow)
+  if (button->default_is_press)
     {
       if (i == 0)
         i = 1;
@@ -685,7 +760,7 @@ gail_button_action_get_name (AtkAction *action,
 
   button = GAIL_BUTTON (action);
 
-  if (button->is_down_arrow)
+  if (button->default_is_press)
     {
       if (i == 0)
         i = 1;
@@ -735,7 +810,7 @@ gail_button_set_description (AtkAction      *action,
 
   button = GAIL_BUTTON (action);
 
-  if (button->is_down_arrow)
+  if (button->default_is_press)
     {
       if (i == 0)
         i = 1;
